/*	V 0.1
    RootJail: Daemon to run other daemons/processes under a root jail
    Copyright (C) 2001,  Luciano Rocha <strange@cryogen.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdio.h>
#define _GNU_SOURCE
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>

struct proc_info {
	pid_t pid;
	uid_t uid;
	gid_t gid;
	int dir;
	time_t e[5];
	int enabled;
	char *name;
	char *path;
	struct proc_info *next;
} *procs = NULL;

void adderror(struct proc_info *l) {

	l->e[0] = l->e[1];
	l->e[1] = l->e[2];
	l->e[2] = l->e[3];
	l->e[3] = l->e[4];
	l->e[4] = time(0);
	if ((l->e[4] - l->e[0]) < 180) {
		syslog(LOG_ERR, "%s respawning too fast, disabling", l->name);
		l->enabled = 0;
	}
}

int my_open(void) {
	int f;

	f = open("./out", O_WRONLY | O_CREAT | O_APPEND, 0666);
	if (f) {
		if (f != 1) dup2(f, 1);
		if (f != 2) dup2(f, 2);
		if (f > 2) close(f);
	}
	return 0;
}

void do_execute(struct proc_info *p) {
	char *args[] = { NULL, NULL };
	char *env[] = { "HOME=/",
	"PATH=/usr/bin:/bin:/usr/sbin:/sbin:/usr/local/bin:/usr/local/sbin",
	"TERM=linux", NULL };

	args[0] = p->path;
	if (fchdir(p->dir)
		|| chroot(".")
		|| setgid(p->gid)
		|| setuid(p->uid)
		|| my_open()
		|| execve(args[0], args, env))
		syslog(LOG_ERR, "%s (%s): %m", p->name, args[0]);
}

inline void execute(struct proc_info *l) {

	if (l->pid != -1 || !l->enabled)
		return;
	if ((l->pid = fork()) == 0) {
		do_execute(l);
		_exit(1);
	} else if (l->pid == -1) {
		adderror(l);
		alarm(5);
	}
}

void sighdl(int s) {
	volatile int e = 0;
	pid_t pid;
	struct proc_info *l;

	signal(s, sighdl);
	switch (s) {
		case SIGHUP:
			syslog(LOG_WARNING, "Signal SIGHUP received, "
					"re-enabeling services");
			l = procs;
			while (l) {
				if (!l->enabled) {
					l->e[0] = l->e[1] = l->e[2] = 0;
					l->e[3] = l->e[4] = 0;
					l->enabled = 1;
					execute(l);
				}
				l = l->next;
			}
			break;
		case SIGALRM:
			l = procs;
			while (l) {
				execute(l);
				l = l->next;
			}
			break;
		case SIGTERM:
			syslog(LOG_WARNING, "Signal SIGTERM received, "
					"killing children and exiting");
			l = procs;
			while (l) {
				l->enabled = 0;
				l = l->next;
			}
			pid = 0;
			while (l) {
				if (l->pid != -1) {
					kill(l->pid, SIGTERM);
					pid = 1;
				}
				l = l->next;
			}
			if (!pid) exit(0);
			usleep(500000);
			pid = 0;
			l = procs;
			while (l) {
				if (l->pid != -1) {
					kill(l->pid, SIGTERM);
					pid = 1;
				}
				l = l->next;
			}
			if (!pid) exit(0);
			usleep(1000000);
			pid = 0;
			l = procs;
			while (l) {
				if (l->pid != -1) {
					kill(l->pid, SIGKILL);
					pid = 1;
				}
				l = l->next;
			}
			exit(0);
		case SIGILL:
		case SIGSEGV:
			++e;
			if (++e < 4) syslog(LOG_ERR, "Signal SIGSEGV/SIGLL "
					"received");
			else if (e == 4) {
				syslog(LOG_ERR, "Signal SIGSEGV/SIGLL "
						"received, aborting");
				abort();
			} else if (e == 5) abort();
			else exit(2);
			break;
		case SIGCHLD:
			pid = waitpid(-1, NULL, WNOHANG | WUNTRACED);
			if (pid <= 0)
				return;
			l = procs;
			while (l && l->pid != pid)
				l = l->next;
			if (!l || !l->enabled) return;
			adderror(l);
			l->pid = -1;
			execute(l);
			break;
	}
}

int main(int ac, char *av[]) {
	struct proc_info *l;
	struct passwd *pw = NULL;
	struct group *gr;
	struct stat st;
	int i, j, p[3];
	char *uid, *gid, *dir, *prog;

	while (*++av) {
		
		for (i = j = 0; (*av)[i]; ++i)
			if ((*av)[i] == ':' && (i < 1 || (*av)[i-1] != '\\')) {
				if (j < 3) p[j] = i;
				++j;
			}
		if (j > 3) {
			fprintf(stderr, "%s: to much options\n", *av);
			continue;
		}
		l = (struct proc_info *) malloc(sizeof(struct proc_info));
		if (!l) {
			fprintf(stderr, "%s: Not enough memory!\n", *av);
			break;
		}
		l->name = strdup(*av);
		if (!l->name) l->name = *av;
		uid = *av;
		gid = dir = prog = NULL;
		i = 0;
		switch (j) {
			case 3:
				(*av)[p[i]] = '\0';
				gid = *av + p[i] + 1;
				++i;
			case 2:
				(*av)[p[i]] = '\0';
				dir = *av + p[i] + 1;
				++i;
			case 1:
				(*av)[p[i]] = '\0';
				prog = *av + p[i] + 1;
		}
		if (isdigit(uid[0])) {
			l->uid = atoi(uid);
			if (!gid && !dir && !prog && !(pw = getpwuid(l->uid))) {
				fprintf(stderr, "%s: No such user\n", uid);
				free(l);
				continue;
			}
		} else if (!(pw = getpwnam(uid))) {
			fprintf(stderr, "%s: No such user\n", uid);
			free(l);
			continue;
		} else l->uid = pw->pw_uid;
		if (gid && !*gid) gid = NULL;
		if (prog && !*prog) prog = NULL;
		if (dir && !*dir) dir = NULL;
		l->dir = open(dir ? dir : pw->pw_dir, O_RDONLY);
		if (l->dir < 0) {
			perror(dir ? dir :pw->pw_dir);
			free(l);
			continue;
		}
		if (fstat(l->dir, &st) || !S_ISDIR(st.st_mode)) {
			fprintf(stderr, "%s: Not a directory: %s\n", *av,
					pw->pw_dir);
			close(l->dir);
			free(l);
			continue;
		}
		if (prog) l->path = prog;
		else {
			i = strlen(pw->pw_shell);
			l->path = NULL;
			if (i != 0 && !(i > 3 && pw->pw_shell[i-1] == 'h'
				&& pw->pw_shell[i-2] == 's'
				&& pw->pw_shell[i-3] != '.'))
				l->path = strdup(pw->pw_shell);
			if (!l->path) l->path = "/go";
		}
		if (gid) {
			if (isdigit(gid[0])) 
				l->gid = atoi(gid);
			else if (!(gr = getgrnam(gid))) {
				fprintf(stderr, "%s: No such group\n", gid);
				free(l);
				continue;
			} else l->gid = gr->gr_gid;
		} else l->gid = pw->pw_gid;
		l->pid = -1;
		l->enabled = 1;
		l->e[0] = l->e[1] = l->e[2] = l->e[3] = l->e[4] = 0;
		l->next = procs;
		procs = l;
	}
	if (!procs) {
		fprintf(stderr, "Nothing to do...\n");
		return 1;
	}
	daemon(0,0);
	openlog("rootjail", LOG_PID, LOG_DAEMON);
	signal(SIGILL, sighdl);
	signal(SIGSEGV, sighdl);
	signal(SIGALRM, sighdl);
	signal(SIGTERM, sighdl);
	signal(SIGCHLD, sighdl);
	signal(SIGHUP, sighdl);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);
	signal(SIGUSR2, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	l = procs;
	while (l) {
		execute(l);
		l = l->next;
	}
	while (1) pause();
	return 0;
}
